from inewave.config import MAX_CORTES, MAX_REES

from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class EstadosPeriodoNwlistcf(Block):
    """
    Bloco do arquivo estados.rel que armazena os estados visitados
    por período na construção dos cortes da FCF.
    """

    BEGIN_PATTERN = "PERIODO: "
    END_PATTERN = "PERIODO: "

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_periodo = Line([IntegerField(4, 19)])
        campos_iniciais: List[Field] = [
            IntegerField(8, 2),
            IntegerField(4, 11),
            IntegerField(4, 16),
            IntegerField(4, 21),
            IntegerField(4, 26),
            FloatField(17, 31, 4),
        ]
        campos_pis: List[Field] = [
            FloatField(17, 49 + 18 * i, 9) for i in range(18)
        ]
        self.__linha = Line(campos_iniciais + campos_pis)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, EstadosPeriodoNwlistcf):
            return False
        bloco: EstadosPeriodoNwlistcf = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            cols = (
                ["IREG", "ITEc", "SIMc", "ITEf", "REE", "FUNC. OBJ.", "EARM"]
                + [f"EAF({i})" for i in range(1, 7)]
                + [f"SGT(P{i}E{j})" for i in range(1, 4) for j in range(1, 4)]
                + ["MX_SAR", "MX_CURVA"]
            )
            df = pd.DataFrame(tabela, columns=cols)
            df = df.astype(
                {
                    "IREG": "int64",
                    "ITEc": "int64",
                    "SIMc": "int64",
                    "ITEf": "int64",
                    "REE": "int64",
                }
            )
            df["PERIODO"] = self.__periodo
            df = df[["PERIODO"] + cols]
            return df

        # Lê o período e as linhas de cabeçalho
        self.__periodo = self.__linha_periodo.read(file.readline())[0]
        for _ in range(2):
            file.readline()

        # Lê as linhas de cortes
        self.__ireg_atual = 0
        self.__itec_atual = 0
        self.__simc_atual = 0
        self.__itef_atual = 0
        tabela = np.zeros((MAX_CORTES * MAX_REES, 24))
        i = 0
        while True:
            ultima_posicao = file.tell()
            linha = file.readline()
            if self.ends(linha) or len(linha) < 3:
                file.seek(ultima_posicao)
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            if dados[0] is not None:
                self.__ireg_atual = dados[0]
            if dados[1] is not None:
                self.__itec_atual = dados[1]
            if dados[2] is not None:
                self.__simc_atual = dados[2]
            if dados[3] is not None:
                self.__itef_atual = dados[3]
            tabela[i, 0] = self.__ireg_atual
            tabela[i, 1] = self.__itec_atual
            tabela[i, 2] = self.__simc_atual
            tabela[i, 3] = self.__itef_atual
            tabela[i, 4:] = dados[4:]
            i += 1



from inewave.newave.modelos.sistema import (
    BlocoCustosDeficit,
    BlocoGeracaoUsinasNaoSimuladas,
    BlocoIntercambioSubsistema,
    BlocoMercadoEnergiaSistema,
    BlocoNumeroPatamaresDeficit,
)


from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class Sistema(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações
    dos subsistemas (submercados).

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoNumeroPatamaresDeficit,
        BlocoCustosDeficit,
        BlocoIntercambioSubsistema,
        BlocoMercadoEnergiaSistema,
        BlocoGeracaoUsinasNaoSimuladas,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="sistema.dat"
    ) -> "Sistema":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="sistema.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def numero_patamares_deficit(self) -> Optional[int]:
        """
        O número de patamares de déficit utilizados no estudo.

        :return: O número de patamares como um inteiro
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoNumeroPatamaresDeficit, 0)
        if b is not None:
            return b.data
        return None

    @numero_patamares_deficit.setter
    def numero_patamares_deficit(self, n: int):
        b = self.__bloco_por_tipo(BlocoNumeroPatamaresDeficit, 0)
        if b is not None:
            b.data = n

    @property
    def custo_deficit(self) -> Optional[pd.DataFrame]:
        """
        Tabela com o custo de cada patamar de déficit, por
        subsistema.

        - Num. Subsistema (`int`)
        - Nome (`str`)
        - Fictício (`int`)
        - Custo Pat 1 (`float`)
        - ...
        - Custo Pat 5 (`float`)
        - Corte Pat 1 (`float`)
        - ...
        - Corte Pat 5 (`float`)

        :return: A duração por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoCustosDeficit, 0)
        if b is not None:
            return b.data
        return None

    @custo_deficit.setter
    def custo_deficit(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoCustosDeficit, 0)
        if b is not None:
            b.data = df

    @property
    def limites_intercambio(self) -> Optional[pd.DataFrame]:
        """
        Tabela com o limite de intercâmbio por par de
        subsistemas.

        - Subsistema De (`int`)
        - Subsistema Para (`int`)
        - Sentido (`int`)
        - Ano (`int`)
        - Janeiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A duração por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoIntercambioSubsistema, 0)
        if b is not None:
            return b.data
        return None

    @limites_intercambio.setter
    def limites_intercambio(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoIntercambioSubsistema, 0)
        if b is not None:
            b.data = df

    @property
    def mercado_energia(self) -> Optional[pd.DataFrame]:
        """
        Tabela com o mercado total de energia por período de estudo.

        - Subsistema (`int`)
        - Ano (`int`)
        - Janeiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoMercadoEnergiaSistema, 0)
        if b is not None:
            return b.data
        return None

    @mercado_energia.setter
    def mercado_energia(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoMercadoEnergiaSistema, 0)
        if b is not None:
            b.data = df

    @property
    def geracao_usinas_nao_simuladas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a geração das usinas não simuladas por fonte
        de geração.

        - Subsistema (`int`)
        - Bloco (`int`)
        - Razão (`str`)
        - Ano (`int`)
        - Janeiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoGeracaoUsinasNaoSimuladas, 0)
        if b is not None:
            return b.data
        return None

    @geracao_usinas_nao_simuladas.setter
    def geracao_usinas_nao_simuladas(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoGeracaoUsinasNaoSimuladas, 0)
        if b is not None:
            b.data = df



from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional, List


from inewave.newave.modelos.dger import BlocoMesInicioEstudo, BlocoNomeCaso
from inewave.newave.modelos.dger import BlocoTipoExecucao
from inewave.newave.modelos.dger import BlocoDuracaoPeriodo
from inewave.newave.modelos.dger import BlocoNumAnosEstudo
from inewave.newave.modelos.dger import BlocoMesInicioPreEstudo
from inewave.newave.modelos.dger import BlocoAnoInicioEstudo
from inewave.newave.modelos.dger import BlocoNumAnosPreEstudo
from inewave.newave.modelos.dger import BlocoNumAnosPosEstudo
from inewave.newave.modelos.dger import BlocoNumAnosPosEstudoSimFinal
from inewave.newave.modelos.dger import BlocoImprimeDados
from inewave.newave.modelos.dger import BlocoImprimeMercados
from inewave.newave.modelos.dger import BlocoImprimeEnergias
from inewave.newave.modelos.dger import BlocoImprimeModeloEstocastico
from inewave.newave.modelos.dger import BlocoImprimeSubsistema
from inewave.newave.modelos.dger import BlocoNumMaxIteracoes
from inewave.newave.modelos.dger import BlocoNumForwards
from inewave.newave.modelos.dger import BlocoNumAberturas
from inewave.newave.modelos.dger import BlocoNumSeriesSinteticas
from inewave.newave.modelos.dger import BlocoOrdemMaximaPARp
from inewave.newave.modelos.dger import BlocoAnoInicialHistorico
from inewave.newave.modelos.dger import BlocoCalculaVolInicial
from inewave.newave.modelos.dger import BlocoVolInicialSubsistema
from inewave.newave.modelos.dger import BlocoTolerancia
from inewave.newave.modelos.dger import BlocoTaxaDesconto
from inewave.newave.modelos.dger import BlocoTipoSimFinal
from inewave.newave.modelos.dger import BlocoImpressaoOperacao
from inewave.newave.modelos.dger import BlocoImpressaoConvergencia
from inewave.newave.modelos.dger import BlocoIntervaloGravar
from inewave.newave.modelos.dger import BlocoMinIteracoes
from inewave.newave.modelos.dger import BlocoRacionamentoPreventivo
from inewave.newave.modelos.dger import BlocoNumAnosManutUTE
from inewave.newave.modelos.dger import BlocoTendenciaHidrologica
from inewave.newave.modelos.dger import BlocoRestricaoItaipu
from inewave.newave.modelos.dger import BlocoBid
from inewave.newave.modelos.dger import BlocoPerdasTransmissao
from inewave.newave.modelos.dger import BlocoElNino
from inewave.newave.modelos.dger import BlocoEnso
from inewave.newave.modelos.dger import BlocoDuracaoPorPatamar
from inewave.newave.modelos.dger import BlocoOutrosUsosAgua
from inewave.newave.modelos.dger import BlocoCorrecaoDesvio
from inewave.newave.modelos.dger import BlocoCurvaAversao
from inewave.newave.modelos.dger import BlocoTipoGeracaoENA
from inewave.newave.modelos.dger import BlocoRiscoDeficit
from inewave.newave.modelos.dger import BlocoIteracaoParaSimFinal
from inewave.newave.modelos.dger import BlocoAgrupamentoLivre
from inewave.newave.modelos.dger import BlocoEqualizacaoPenalInt
from inewave.newave.modelos.dger import BlocoRepresentacaoSubmot
from inewave.newave.modelos.dger import BlocoOrdenacaoAutomatica
from inewave.newave.modelos.dger import BlocoConsideraCargaAdicional
from inewave.newave.modelos.dger import BlocoDeltaZSUP
from inewave.newave.modelos.dger import BlocoDeltaZINF
from inewave.newave.modelos.dger import BlocoDeltasConsecutivos
from inewave.newave.modelos.dger import BlocoDespachoAntecipadoGNL
from inewave.newave.modelos.dger import BlocoModifAutomaticaAdTerm
from inewave.newave.modelos.dger import BlocoGeracaoHidraulicaMin
from inewave.newave.modelos.dger import BlocoSimFinalComData
from inewave.newave.modelos.dger import BlocoGerenciamentoPLs
from inewave.newave.modelos.dger import BlocoSAR
from inewave.newave.modelos.dger import BlocoCVAR
from inewave.newave.modelos.dger import BlocoZSUPMinConvergencia
from inewave.newave.modelos.dger import BlocoDesconsideraVazaoMinima
from inewave.newave.modelos.dger import BlocoRestricoesEletricas
from inewave.newave.modelos.dger import BlocoSelecaoCortes
from inewave.newave.modelos.dger import BlocoJanelaCortes
from inewave.newave.modelos.dger import BlocoReamostragemCenarios
from inewave.newave.modelos.dger import BlocoConvergeNoZero
from inewave.newave.modelos.dger import BlocoConsultaFCF
from inewave.newave.modelos.dger import BlocoImpressaoENA
from inewave.newave.modelos.dger import BlocoImpressaoCortesAtivosSimFinal
from inewave.newave.modelos.dger import BlocoRepresentacaoAgregacao
from inewave.newave.modelos.dger import BlocoMatrizCorrelacaoEspacial
from inewave.newave.modelos.dger import BlocoDesconsideraConvEstatistica
from inewave.newave.modelos.dger import BlocoMomentoReamostragem
from inewave.newave.modelos.dger import BlocoMantemArquivosEnergias
from inewave.newave.modelos.dger import BlocoInicioTesteConvergencia
from inewave.newave.modelos.dger import BlocoSazonalizarVminT
from inewave.newave.modelos.dger import BlocoSazonalizarVmaxT
from inewave.newave.modelos.dger import BlocoSazonalizarVminP
from inewave.newave.modelos.dger import BlocoSazonalizarCfugaCmont
from inewave.newave.modelos.dger import BlocoRestricoesEmissaoGEE
from inewave.newave.modelos.dger import BlocoAfluenciaAnualPARp
from inewave.newave.modelos.dger import BlocoRestricoesFornecGas
from inewave.newave.modelos.dger import BlocoMemCalculoCortes
from inewave.newave.modelos.dger import BlocoGeracaoEolica
from inewave.newave.modelos.dger import BlocoCompensacaoCorrelacaoCruzada
from inewave.newave.modelos.dger import (
    BlocoConsideracaoTurbinamentoMinimoMaximo,
)
from inewave.newave.modelos.dger import BlocoConsideracaoDefluenciaMaxima
from inewave.newave.modelos.dger import BlocoAproveitamentoBasePLsBackward
from inewave.newave.modelos.dger import BlocoImpressaoEstadosGeracaoCortes
from inewave.newave.modelos.dger import BlocoSementeForward
from inewave.newave.modelos.dger import BlocoSementeBackward
from inewave.newave.modelos.dger import BlocoRestricaoLPPTurbinamentoMaximoREE
from inewave.newave.modelos.dger import BlocoRestricaoLPPDefluenciaMaximaREE
from inewave.newave.modelos.dger import BlocoRestricaoLPPTurbinamentoMaximoUHE
from inewave.newave.modelos.dger import BlocoRestricaoLPPDefluenciaMaximaUHE


class DGer(SectionFile):
    """
    Classe para armazenar dados gerais de uma execução do NEWAVE.

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoNomeCaso,
        BlocoTipoExecucao,
        BlocoDuracaoPeriodo,
        BlocoNumAnosEstudo,
        BlocoMesInicioPreEstudo,
        BlocoMesInicioEstudo,
        BlocoAnoInicioEstudo,
        BlocoNumAnosPreEstudo,
        BlocoNumAnosPosEstudo,
        BlocoNumAnosPosEstudoSimFinal,
        BlocoImprimeDados,
        BlocoImprimeMercados,
        BlocoImprimeEnergias,
        BlocoImprimeModeloEstocastico,
        BlocoImprimeSubsistema,
        BlocoNumMaxIteracoes,
        BlocoNumForwards,
        BlocoNumAberturas,
        BlocoNumSeriesSinteticas,
        BlocoOrdemMaximaPARp,
        BlocoAnoInicialHistorico,
        BlocoCalculaVolInicial,
        BlocoVolInicialSubsistema,
        BlocoTolerancia,
        BlocoTaxaDesconto,
        BlocoTipoSimFinal,
        BlocoImpressaoOperacao,
        BlocoImpressaoConvergencia,
        BlocoIntervaloGravar,
        BlocoMinIteracoes,
        BlocoRacionamentoPreventivo,
        BlocoNumAnosManutUTE,
        BlocoTendenciaHidrologica,
        BlocoRestricaoItaipu,
        BlocoBid,
        BlocoPerdasTransmissao,
        BlocoElNino,
        BlocoEnso,
        BlocoDuracaoPorPatamar,
        BlocoOutrosUsosAgua,
        BlocoCorrecaoDesvio,
        BlocoCurvaAversao,
        BlocoTipoGeracaoENA,
        BlocoRiscoDeficit,
        BlocoIteracaoParaSimFinal,
        BlocoAgrupamentoLivre,
        BlocoEqualizacaoPenalInt,
        BlocoRepresentacaoSubmot,
        BlocoOrdenacaoAutomatica,
        BlocoConsideraCargaAdicional,
        BlocoDeltaZSUP,
        BlocoDeltaZINF,
        BlocoDeltasConsecutivos,
        BlocoDespachoAntecipadoGNL,
        BlocoModifAutomaticaAdTerm,
        BlocoGeracaoHidraulicaMin,
        BlocoSimFinalComData,
        BlocoGerenciamentoPLs,
        BlocoSAR,
        BlocoCVAR,
        BlocoZSUPMinConvergencia,
        BlocoDesconsideraVazaoMinima,
        BlocoRestricoesEletricas,
        BlocoSelecaoCortes,
        BlocoJanelaCortes,
        BlocoReamostragemCenarios,
        BlocoConvergeNoZero,
        BlocoConsultaFCF,
        BlocoImpressaoENA,
        BlocoImpressaoCortesAtivosSimFinal,
        BlocoRepresentacaoAgregacao,
        BlocoMatrizCorrelacaoEspacial,
        BlocoDesconsideraConvEstatistica,
        BlocoMomentoReamostragem,
        BlocoMantemArquivosEnergias,
        BlocoInicioTesteConvergencia,
        BlocoSazonalizarVminT,
        BlocoSazonalizarVmaxT,
        BlocoSazonalizarVminP,
        BlocoSazonalizarCfugaCmont,
        BlocoRestricoesEmissaoGEE,
        BlocoAfluenciaAnualPARp,
        BlocoRestricoesFornecGas,
        BlocoMemCalculoCortes,
        BlocoGeracaoEolica,
        BlocoCompensacaoCorrelacaoCruzada,
        BlocoConsideracaoTurbinamentoMinimoMaximo,
        BlocoConsideracaoDefluenciaMaxima,
        BlocoAproveitamentoBasePLsBackward,
        BlocoImpressaoEstadosGeracaoCortes,
        BlocoSementeForward,
        BlocoSementeBackward,
        BlocoRestricaoLPPTurbinamentoMaximoREE,
        BlocoRestricaoLPPDefluenciaMaximaREE,
        BlocoRestricaoLPPTurbinamentoMaximoUHE,
        BlocoRestricaoLPPDefluenciaMaximaUHE,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="dger.dat") -> "DGer":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="dger.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def nome_caso(self) -> Optional[str]:
        """
        Configuração da linha número 1 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: str
        """
        b = self.__bloco_por_tipo(BlocoNomeCaso, 0)
        if b is not None:
            return b.valor
        return None

    @nome_caso.setter
    def nome_caso(self, dado: str):
        b = self.__bloco_por_tipo(BlocoNomeCaso, 0)
        if b is not None:
            b.valor = dado

    @property
    def tipo_execucao(self) -> Optional[int]:
        """
        Configuração da linha número 2 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoTipoExecucao, 0)
        if b is not None:
            return b.valor
        return None

    @tipo_execucao.setter
    def tipo_execucao(self, dado: int):
        b = self.__bloco_por_tipo(BlocoTipoExecucao, 0)
        if b is not None:
            b.valor = dado

    @property
    def duracao_periodo(self) -> Optional[int]:
        """
        Configuração da linha número 3 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoDuracaoPeriodo, 0)
        if b is not None:
            return b.valor
        return None

    @duracao_periodo.setter
    def duracao_periodo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoDuracaoPeriodo, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_anos_estudo(self) -> Optional[int]:
        """
        Configuração da linha número 4 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumAnosEstudo, 0)
        if b is not None:
            return b.valor
        return None

    @num_anos_estudo.setter
    def num_anos_estudo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumAnosEstudo, 0)
        if b is not None:
            b.valor = dado

    @property
    def mes_inicio_pre_estudo(self) -> Optional[int]:
        """
        Configuração da linha número 5 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoMesInicioPreEstudo, 0)
        if b is not None:
            return b.valor
        return None

    @mes_inicio_pre_estudo.setter
    def mes_inicio_pre_estudo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoMesInicioPreEstudo, 0)
        if b is not None:
            b.valor = dado

    @property
    def mes_inicio_estudo(self) -> Optional[int]:
        """
        Configuração da linha número 6 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoMesInicioEstudo, 0)
        if b is not None:
            return b.valor
        return None

    @mes_inicio_estudo.setter
    def mes_inicio_estudo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoMesInicioEstudo, 0)
        if b is not None:
            b.valor = dado

    @property
    def ano_inicio_estudo(self) -> Optional[int]:
        """
        Configuração da linha número 7 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoAnoInicioEstudo, 0)
        if b is not None:
            return b.valor
        return None

    @ano_inicio_estudo.setter
    def ano_inicio_estudo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoAnoInicioEstudo, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_anos_pre_estudo(self) -> Optional[int]:
        """
        Configuração da linha número 8 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumAnosPreEstudo, 0)
        if b is not None:
            return b.valor
        return None

    @num_anos_pre_estudo.setter
    def num_anos_pre_estudo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumAnosPreEstudo, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_anos_pos_estudo(self) -> Optional[int]:
        """
        Configuração da linha número 9 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumAnosPosEstudo, 0)
        if b is not None:
            return b.valor
        return None

    @num_anos_pos_estudo.setter
    def num_anos_pos_estudo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumAnosPosEstudo, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_anos_pos_sim_final(self) -> Optional[int]:
        """
        Configuração da linha número 10 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumAnosPosEstudoSimFinal, 0)
        if b is not None:
            return b.valor
        return None

    @num_anos_pos_sim_final.setter
    def num_anos_pos_sim_final(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumAnosPosEstudoSimFinal, 0)
        if b is not None:
            b.valor = dado

    @property
    def imprime_dados(self) -> Optional[int]:
        """
        Configuração da linha número 11 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImprimeDados, 0)
        if b is not None:
            return b.valor
        return None

    @imprime_dados.setter
    def imprime_dados(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImprimeDados, 0)
        if b is not None:
            b.valor = dado

    @property
    def imprime_mercados(self) -> Optional[int]:
        """
        Configuração da linha número 12 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImprimeMercados, 0)
        if b is not None:
            return b.valor
        return None

    @imprime_mercados.setter
    def imprime_mercados(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImprimeMercados, 0)
        if b is not None:
            b.valor = dado

    @property
    def imprime_energias(self) -> Optional[int]:
        """
        Configuração da linha número 13 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImprimeEnergias, 0)
        if b is not None:
            return b.valor
        return None

    @imprime_energias.setter
    def imprime_energias(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImprimeEnergias, 0)
        if b is not None:
            b.valor = dado

    @property
    def imprime_modelo_estocastico(self) -> Optional[int]:
        """
        Configuração da linha número 14 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImprimeModeloEstocastico, 0)
        if b is not None:
            return b.valor
        return None

    @imprime_modelo_estocastico.setter
    def imprime_modelo_estocastico(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImprimeModeloEstocastico, 0)
        if b is not None:
            b.valor = dado

    @property
    def imprime_subsistema(self) -> Optional[int]:
        """
        Configuração da linha número 15 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImprimeSubsistema, 0)
        if b is not None:
            return b.valor
        return None

    @imprime_subsistema.setter
    def imprime_subsistema(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImprimeSubsistema, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_max_iteracoes(self) -> Optional[int]:
        """
        Configuração da linha número 16 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumMaxIteracoes, 0)
        if b is not None:
            return b.valor
        return None

    @num_max_iteracoes.setter
    def num_max_iteracoes(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumMaxIteracoes, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_forwards(self) -> Optional[int]:
        """
        Configuração da linha número 17 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumForwards, 0)
        if b is not None:
            return b.valor
        return None

    @num_forwards.setter
    def num_forwards(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumForwards, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_aberturas(self) -> Optional[int]:
        """
        Configuração da linha número 18 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumAberturas, 0)
        if b is not None:
            return b.valor
        return None

    @num_aberturas.setter
    def num_aberturas(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumAberturas, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_series_sinteticas(self) -> Optional[int]:
        """
        Configuração da linha número 19 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumSeriesSinteticas, 0)
        if b is not None:
            return b.valor
        return None

    @num_series_sinteticas.setter
    def num_series_sinteticas(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumSeriesSinteticas, 0)
        if b is not None:
            b.valor = dado

    @property
    def ordem_maxima_parp(self) -> Optional[int]:
        """
        Configuração da linha número 20 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoOrdemMaximaPARp, 0)
        if b is not None:
            return b.valor
        return None

    @ordem_maxima_parp.setter
    def ordem_maxima_parp(self, dado: int):
        b = self.__bloco_por_tipo(BlocoOrdemMaximaPARp, 0)
        if b is not None:
            b.valor = dado

    @property
    def ano_inicial_historico(self) -> Optional[int]:
        """
        Configuração da linha número 21 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoAnoInicialHistorico, 0)
        if b is not None:
            return b.ano_inicial
        return None

    @ano_inicial_historico.setter
    def ano_inicial_historico(self, dado: int):
        b = self.__bloco_por_tipo(BlocoAnoInicialHistorico, 0)
        if b is not None:
            b.ano_inicial = dado

    @property
    def tamanho_registro_arquivo_historico(self) -> Optional[int]:
        """
        Configuração da linha número 21 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoAnoInicialHistorico, 0)
        if b is not None:
            return b.tamanho_registro_arquivo
        return None

    @tamanho_registro_arquivo_historico.setter
    def tamanho_registro_arquivo_historico(self, dado: int):
        b = self.__bloco_por_tipo(BlocoAnoInicialHistorico, 0)
        if b is not None:
            b.tamanho_registro_arquivo = dado

    @property
    def calcula_volume_inicial(self) -> Optional[int]:
        """
        Configuração da linha número 22 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoCalculaVolInicial, 0)
        if b is not None:
            return b.valor
        return None

    @calcula_volume_inicial.setter
    def calcula_volume_inicial(self, dado: int):
        b = self.__bloco_por_tipo(BlocoCalculaVolInicial, 0)
        if b is not None:
            b.valor = dado

    @property
    def volume_inicial_subsistema(self) -> List[Optional[float]]:
        """
        Configuração das linhas número 23 e 24 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoVolInicialSubsistema, 0)
        if b is not None:
            return b.valores
        return []

    @volume_inicial_subsistema.setter
    def volume_inicial_subsistema(self, dado: List[Optional[float]]):
        b = self.__bloco_por_tipo(BlocoVolInicialSubsistema, 0)
        if b is not None:
            b.valores = dado

    @property
    def tolerancia(self) -> Optional[float]:
        """
        Configuração da linha número 25 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoTolerancia, 0)
        if b is not None:
            return b.valor
        return None

    @tolerancia.setter
    def tolerancia(self, dado: float):
        b = self.__bloco_por_tipo(BlocoTolerancia, 0)
        if b is not None:
            b.valor = dado

    @property
    def taxa_de_desconto(self) -> Optional[float]:
        """
        Configuração da linha número 26 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoTaxaDesconto, 0)
        if b is not None:
            return b.valor
        return None

    @taxa_de_desconto.setter
    def taxa_de_desconto(self, dado: float):
        b = self.__bloco_por_tipo(BlocoTaxaDesconto, 0)
        if b is not None:
            b.valor = dado

    @property
    def tipo_simulacao_final(self) -> Optional[int]:
        """
        Configuração do primeiro campo da linha número 27
        do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoTipoSimFinal, 0)
        if b is not None:
            return b.valor[0]
        return None

    @tipo_simulacao_final.setter
    def tipo_simulacao_final(self, dado: int):
        b = self.__bloco_por_tipo(BlocoTipoSimFinal, 0)
        if b is not None:
            b.valor = [dado] + [b.valor[1]]

    @property
    def agregacao_simulacao_final(self) -> Optional[int]:
        """
        Configuração do segundo campo da linha número 27
        do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoTipoSimFinal, 0)
        if b is not None:
            return b.valor[1]
        return None

    @agregacao_simulacao_final.setter
    def agregacao_simulacao_final(self, dado: int):
        b = self.__bloco_por_tipo(BlocoTipoSimFinal, 0)
        if b is not None:
            b.valor = [b.valor[0]] + [dado]

    @property
    def impressao_operacao(self) -> Optional[int]:
        """
        Configuração da linha número 28 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImpressaoOperacao, 0)
        if b is not None:
            return b.valor
        return None

    @impressao_operacao.setter
    def impressao_operacao(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImpressaoOperacao, 0)
        if b is not None:
            b.valor = dado

    @property
    def impressao_convergencia(self) -> Optional[int]:
        """
        Configuração da linha número 29 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImpressaoConvergencia, 0)
        if b is not None:
            return b.valor
        return None

    @impressao_convergencia.setter
    def impressao_convergencia(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImpressaoConvergencia, 0)
        if b is not None:
            b.valor = dado

    @property
    def intervalo_para_gravar(self) -> Optional[int]:
        """
        Configuração da linha número 30 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoIntervaloGravar, 0)
        if b is not None:
            return b.valor
        return None

    @intervalo_para_gravar.setter
    def intervalo_para_gravar(self, dado: int):
        b = self.__bloco_por_tipo(BlocoIntervaloGravar, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_minimo_iteracoes(self) -> Optional[int]:
        """
        Configuração da linha número 31 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoMinIteracoes, 0)
        if b is not None:
            return b.valor
        return None

    @num_minimo_iteracoes.setter
    def num_minimo_iteracoes(self, dado: int):
        b = self.__bloco_por_tipo(BlocoMinIteracoes, 0)
        if b is not None:
            b.valor = dado

    @property
    def racionamento_preventivo(self) -> Optional[int]:
        """
        Configuração da linha número 32 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoRacionamentoPreventivo, 0)
        if b is not None:
            return b.valor
        return None

    @racionamento_preventivo.setter
    def racionamento_preventivo(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRacionamentoPreventivo, 0)
        if b is not None:
            b.valor = dado

    @property
    def num_anos_manutencao_utes(self) -> Optional[int]:
        """
        Configuração da linha número 33 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoNumAnosManutUTE, 0)
        if b is not None:
            return b.valor
        return None

    @num_anos_manutencao_utes.setter
    def num_anos_manutencao_utes(self, dado: int):
        b = self.__bloco_por_tipo(BlocoNumAnosManutUTE, 0)
        if b is not None:
            b.valor = dado

    @property
    def considera_tendencia_hidrologica_calculo_politica(
        self,
    ) -> Optional[int]:
        """
        Configuração da linha número 34 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoTendenciaHidrologica, 0)
        if b is not None:
            return b.considera_tendencia_hidrologica_calculo_politica
        return None

    @considera_tendencia_hidrologica_calculo_politica.setter
    def considera_tendencia_hidrologica_calculo_politica(self, dado: int):
        b = self.__bloco_por_tipo(BlocoTendenciaHidrologica, 0)
        if b is not None:
            b.considera_tendencia_hidrologica_calculo_politica = dado

    @property
    def considera_tendencia_hidrologica_sim_final(self) -> Optional[int]:
        """
        Configuração da linha número 34 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoTendenciaHidrologica, 0)
        if b is not None:
            return b.considera_tendencia_hidrologica_sim_final
        return None

    @considera_tendencia_hidrologica_sim_final.setter
    def considera_tendencia_hidrologica_sim_final(self, dado: int):
        b = self.__bloco_por_tipo(BlocoTendenciaHidrologica, 0)
        if b is not None:
            b.considera_tendencia_hidrologica_sim_final = dado

    @property
    def restricao_itaipu(self) -> Optional[int]:
        """
        Configuração da linha número 35 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoRestricaoItaipu, 0)
        if b is not None:
            return b.valor
        return None

    @restricao_itaipu.setter
    def restricao_itaipu(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricaoItaipu, 0)
        if b is not None:
            b.valor = dado

    @property
    def bid(self) -> Optional[int]:
        """
        Configuração da linha número 36 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoBid, 0)
        if b is not None:
            return b.valor
        return None

    @bid.setter
    def bid(self, dado: int):
        b = self.__bloco_por_tipo(BlocoBid, 0)
        if b is not None:
            b.valor = dado

    @property
    def perdas_rede_transmissao(self) -> Optional[int]:
        """
        Configuração da linha número 37 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoPerdasTransmissao, 0)
        if b is not None:
            return b.valor
        return None

    @perdas_rede_transmissao.setter
    def perdas_rede_transmissao(self, dado: int):
        b = self.__bloco_por_tipo(BlocoPerdasTransmissao, 0)
        if b is not None:
            b.valor = dado

    @property
    def el_nino(self) -> Optional[int]:
        """
        Configuração da linha número 38 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoElNino, 0)
        if b is not None:
            return b.valor
        return None

    @el_nino.setter
    def el_nino(self, dado: int):
        b = self.__bloco_por_tipo(BlocoElNino, 0)
        if b is not None:
            b.valor = dado

    @property
    def enso(self) -> Optional[int]:
        """
        Configuração da linha número 39 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoEnso, 0)
        if b is not None:
            return b.valor
        return None

    @enso.setter
    def enso(self, dado: int):
        b = self.__bloco_por_tipo(BlocoEnso, 0)
        if b is not None:
            b.valor = dado

    @property
    def duracao_por_patamar(self) -> Optional[int]:
        """
        Configuração da linha número 40 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoDuracaoPorPatamar, 0)
        if b is not None:
            return b.valor
        return None

    @duracao_por_patamar.setter
    def duracao_por_patamar(self, dado: int):
        b = self.__bloco_por_tipo(BlocoDuracaoPorPatamar, 0)
        if b is not None:
            b.valor = dado

    @property
    def outros_usos_da_agua(self) -> Optional[int]:
        """
        Configuração da linha número 41 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoOutrosUsosAgua, 0)
        if b is not None:
            return b.valor
        return None

    @outros_usos_da_agua.setter
    def outros_usos_da_agua(self, dado: int):
        b = self.__bloco_por_tipo(BlocoOutrosUsosAgua, 0)
        if b is not None:
            b.valor = dado

    @property
    def correcao_desvio(self) -> Optional[int]:
        """
        Configuração da linha número 42 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoCorrecaoDesvio, 0)
        if b is not None:
            return b.valor
        return None

    @correcao_desvio.setter
    def correcao_desvio(self, dado: int):
        b = self.__bloco_por_tipo(BlocoCorrecaoDesvio, 0)
        if b is not None:
            b.valor = dado

    @property
    def curva_aversao(self) -> Optional[int]:
        """
        Configuração da linha número 43 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoCurvaAversao, 0)
        if b is not None:
            return b.valor
        return None

    @curva_aversao.setter
    def curva_aversao(self, dado: int):
        b = self.__bloco_por_tipo(BlocoCurvaAversao, 0)
        if b is not None:
            b.valor = dado

    @property
    def tipo_geracao_enas(self) -> Optional[int]:
        """
        Configuração da linha número 44 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoTipoGeracaoENA, 0)
        if b is not None:
            return b.valor
        return None

    @tipo_geracao_enas.setter
    def tipo_geracao_enas(self, dado: int):
        b = self.__bloco_por_tipo(BlocoTipoGeracaoENA, 0)
        if b is not None:
            b.valor = dado

    @property
    def primeira_profundidade_risco_deficit(self) -> Optional[float]:
        """
        Configuração da linha número 45 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: float
        """
        b = self.__bloco_por_tipo(BlocoRiscoDeficit, 0)
        if b is not None:
            return b.primeira_profundidade_risco_deficit
        return None

    @primeira_profundidade_risco_deficit.setter
    def primeira_profundidade_risco_deficit(self, dado: float):
        b = self.__bloco_por_tipo(BlocoRiscoDeficit, 0)
        if b is not None:
            b.primeira_profundidade_risco_deficit = dado

    @property
    def segunda_profundidade_risco_deficit(self) -> Optional[float]:
        """
        Configuração da linha número 45 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: float
        """
        b = self.__bloco_por_tipo(BlocoRiscoDeficit, 0)
        if b is not None:
            return b.segunda_profundidade_risco_deficit
        return None

    @segunda_profundidade_risco_deficit.setter
    def segunda_profundidade_risco_deficit(self, dado: float):
        b = self.__bloco_por_tipo(BlocoRiscoDeficit, 0)
        if b is not None:
            b.segunda_profundidade_risco_deficit = dado

    @property
    def iteracao_para_simulacao_final(self) -> Optional[int]:
        """
        Configuração da linha número 46 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoIteracaoParaSimFinal, 0)
        if b is not None:
            return b.valor
        return None

    @iteracao_para_simulacao_final.setter
    def iteracao_para_simulacao_final(self, dado: int):
        b = self.__bloco_por_tipo(BlocoIteracaoParaSimFinal, 0)
        if b is not None:
            b.valor = dado

    @property
    def agrupamento_livre(self) -> Optional[int]:
        """
        Configuração da linha número 47 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoAgrupamentoLivre, 0)
        if b is not None:
            return b.valor
        return None

    @agrupamento_livre.setter
    def agrupamento_livre(self, dado: int):
        b = self.__bloco_por_tipo(BlocoAgrupamentoLivre, 0)
        if b is not None:
            b.valor = dado

    @property
    def equalizacao_penal_intercambio(self) -> Optional[int]:
        """
        Configuração da linha número 48 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoEqualizacaoPenalInt, 0)
        if b is not None:
            return b.valor
        return None

    @equalizacao_penal_intercambio.setter
    def equalizacao_penal_intercambio(self, dado: int):
        b = self.__bloco_por_tipo(BlocoEqualizacaoPenalInt, 0)
        if b is not None:
            b.valor = dado

    @property
    def representacao_submotorizacao(self) -> Optional[int]:
        """
        Configuração da linha número 49 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoRepresentacaoSubmot, 0)
        if b is not None:
            return b.valor
        return None

    @representacao_submotorizacao.setter
    def representacao_submotorizacao(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRepresentacaoSubmot, 0)
        if b is not None:
            b.valor = dado

    @property
    def ordenacao_automatica(self) -> Optional[int]:
        """
        Configuração da linha número 50 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoOrdenacaoAutomatica, 0)
        if b is not None:
            return b.valor
        return None

    @ordenacao_automatica.setter
    def ordenacao_automatica(self, dado: int):
        b = self.__bloco_por_tipo(BlocoOrdenacaoAutomatica, 0)
        if b is not None:
            b.valor = dado

    @property
    def considera_carga_adicional(self) -> Optional[int]:
        """
        Configuração da linha número 51 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoConsideraCargaAdicional, 0)
        if b is not None:
            return b.valor
        return None

    @considera_carga_adicional.setter
    def considera_carga_adicional(self, dado: int):
        b = self.__bloco_por_tipo(BlocoConsideraCargaAdicional, 0)
        if b is not None:
            b.valor = dado

    @property
    def delta_zsup(self) -> Optional[float]:
        """
        Configuração da linha número 52 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: float
        """
        b = self.__bloco_por_tipo(BlocoDeltaZSUP, 0)
        if b is not None:
            return b.valor
        return None

    @delta_zsup.setter
    def delta_zsup(self, dado: float):
        b = self.__bloco_por_tipo(BlocoDeltaZSUP, 0)
        if b is not None:
            b.valor = dado

    @property
    def delta_zinf(self) -> Optional[float]:
        """
        Configuração da linha número 53 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: float
        """
        b = self.__bloco_por_tipo(BlocoDeltaZINF, 0)
        if b is not None:
            return b.valor
        return None

    @delta_zinf.setter
    def delta_zinf(self, dado: float):
        b = self.__bloco_por_tipo(BlocoDeltaZINF, 0)
        if b is not None:
            b.valor = dado

    @property
    def deltas_consecutivos(self) -> Optional[int]:
        """
        Configuração da linha número 54 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoDeltasConsecutivos, 0)
        if b is not None:
            return b.valor
        return None

    @deltas_consecutivos.setter
    def deltas_consecutivos(self, dado: int):
        b = self.__bloco_por_tipo(BlocoDeltasConsecutivos, 0)
        if b is not None:
            b.valor = dado

    @property
    def despacho_antecipado_gnl(self) -> Optional[int]:
        """
        Configuração da linha número 55 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoDespachoAntecipadoGNL, 0)
        if b is not None:
            return b.valor
        return None

    @despacho_antecipado_gnl.setter
    def despacho_antecipado_gnl(self, dado: int):
        b = self.__bloco_por_tipo(BlocoDespachoAntecipadoGNL, 0)
        if b is not None:
            b.valor = dado

    @property
    def modif_automatica_adterm(self) -> Optional[int]:
        """
        Configuração da linha número 56 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoModifAutomaticaAdTerm, 0)
        if b is not None:
            return b.valor
        return None

    @modif_automatica_adterm.setter
    def modif_automatica_adterm(self, dado: int):
        b = self.__bloco_por_tipo(BlocoModifAutomaticaAdTerm, 0)
        if b is not None:
            b.valor = dado

    @property
    def considera_ghmin(self) -> Optional[int]:
        """
        Configuração da linha número 57 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGeracaoHidraulicaMin, 0)
        if b is not None:
            return b.valor
        return None

    @considera_ghmin.setter
    def considera_ghmin(self, dado: int):
        b = self.__bloco_por_tipo(BlocoGeracaoHidraulicaMin, 0)
        if b is not None:
            b.valor = dado

    @property
    def simulacao_final_com_data(self) -> Optional[int]:
        """
        Configuração da linha número 58 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSimFinalComData, 0)
        if b is not None:
            return b.valor
        return None

    @simulacao_final_com_data.setter
    def simulacao_final_com_data(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSimFinalComData, 0)
        if b is not None:
            b.valor = dado

    @property
    def utiliza_gerenciamento_pls(self) -> Optional[int]:
        """
        Configuração da linha número 59 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            return b.utiliza_gerenciamento_pls
        return None

    @utiliza_gerenciamento_pls.setter
    def utiliza_gerenciamento_pls(self, dado: int):
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            b.utiliza_gerenciamento_pls = dado

    @property
    def comunicacao_dois_niveis(self) -> Optional[int]:
        """
        Configuração da linha número 59 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            return b.comunicacao_dois_niveis
        return None

    @comunicacao_dois_niveis.setter
    def comunicacao_dois_niveis(self, dado: int):
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            b.comunicacao_dois_niveis = dado

    @property
    def armazenamento_local_arquivos_temporarios(self) -> Optional[int]:
        """
        Configuração da linha número 59 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            return b.armazenamento_local_arquivos_temporarios
        return None

    @armazenamento_local_arquivos_temporarios.setter
    def armazenamento_local_arquivos_temporarios(self, dado: int):
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            b.armazenamento_local_arquivos_temporarios = dado

    @property
    def alocacao_memoria_ena(self) -> Optional[int]:
        """
        Configuração da linha número 59 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            return b.alocacao_memoria_ena
        return None

    @alocacao_memoria_ena.setter
    def alocacao_memoria_ena(self, dado: int):
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            b.alocacao_memoria_ena = dado

    @property
    def alocacao_memoria_cortes(self) -> Optional[int]:
        """
        Configuração da linha número 59 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            return b.alocacao_memoria_cortes
        return None

    @alocacao_memoria_cortes.setter
    def alocacao_memoria_cortes(self, dado: int):
        b = self.__bloco_por_tipo(BlocoGerenciamentoPLs, 0)
        if b is not None:
            b.alocacao_memoria_cortes = dado

    @property
    def sar(self) -> Optional[int]:
        """
        Configuração da linha número 60 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSAR, 0)
        if b is not None:
            return b.valor
        return None

    @sar.setter
    def sar(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSAR, 0)
        if b is not None:
            b.valor = dado

    @property
    def cvar(self) -> Optional[int]:
        """
        Configuração da linha número 61 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoCVAR, 0)
        if b is not None:
            return b.valor
        return None

    @cvar.setter
    def cvar(self, dado: int):
        b = self.__bloco_por_tipo(BlocoCVAR, 0)
        if b is not None:
            b.valor = dado

    @property
    def considera_zsup_min_convergencia(self) -> Optional[int]:
        """
        Configuração da linha número 62 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoZSUPMinConvergencia, 0)
        if b is not None:
            return b.valor
        return None

    @considera_zsup_min_convergencia.setter
    def considera_zsup_min_convergencia(self, dado: int):
        b = self.__bloco_por_tipo(BlocoZSUPMinConvergencia, 0)
        if b is not None:
            b.valor = dado

    @property
    def desconsidera_vazao_minima(self) -> Optional[int]:
        """
        Configuração da linha número 63 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoDesconsideraVazaoMinima, 0)
        if b is not None:
            return b.valor
        return None

    @desconsidera_vazao_minima.setter
    def desconsidera_vazao_minima(self, dado: int):
        b = self.__bloco_por_tipo(BlocoDesconsideraVazaoMinima, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricoes_eletricas(self) -> Optional[int]:
        """
        Configuração da linha número 64 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoRestricoesEletricas, 0)
        if b is not None:
            return b.valor
        return None

    @restricoes_eletricas.setter
    def restricoes_eletricas(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricoesEletricas, 0)
        if b is not None:
            b.valor = dado

    @property
    def selecao_de_cortes_backward(self) -> Optional[int]:
        """
        Configuração do primeiro campo da linha número 65 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSelecaoCortes, 0)
        if b is not None:
            return b.considera_na_backward
        return None

    @selecao_de_cortes_backward.setter
    def selecao_de_cortes_backward(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSelecaoCortes, 0)
        if b is not None:
            b.considera_na_backward = dado

    @property
    def selecao_de_cortes_forward(self) -> Optional[int]:
        """
        Configuração do segundo campo da linha número 65 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSelecaoCortes, 0)
        if b is not None:
            return b.considera_na_forward
        return None

    @selecao_de_cortes_forward.setter
    def selecao_de_cortes_forward(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSelecaoCortes, 0)
        if b is not None:
            b.considera_na_forward = dado

    @property
    def janela_de_cortes(self) -> Optional[int]:
        """
        Configuração da linha número 66 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoJanelaCortes, 0)
        if b is not None:
            return b.valor
        return None

    @janela_de_cortes.setter
    def janela_de_cortes(self, dado: int):
        b = self.__bloco_por_tipo(BlocoJanelaCortes, 0)
        if b is not None:
            b.valor = dado

    @property
    def considera_reamostragem_cenarios(self) -> Optional[int]:
        """
        Configuração da linha número 67 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoReamostragemCenarios, 0)
        if b is not None:
            return b.considera_reamostragem_cenarios
        return None

    @considera_reamostragem_cenarios.setter
    def considera_reamostragem_cenarios(self, dado: int):
        b = self.__bloco_por_tipo(BlocoReamostragemCenarios, 0)
        if b is not None:
            b.considera_reamostragem_cenarios = dado

    @property
    def tipo_reamostragem_cenarios(self) -> Optional[int]:
        """
        Configuração da linha número 67 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoReamostragemCenarios, 0)
        if b is not None:
            return b.tipo_reamostragem_cenarios
        return None

    @tipo_reamostragem_cenarios.setter
    def tipo_reamostragem_cenarios(self, dado: int):
        b = self.__bloco_por_tipo(BlocoReamostragemCenarios, 0)
        if b is not None:
            b.tipo_reamostragem_cenarios = dado

    @property
    def passo_reamostragem_cenarios(self) -> Optional[int]:
        """
        Configuração da linha número 67 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoReamostragemCenarios, 0)
        if b is not None:
            return b.passo_reamostragem_cenarios
        return None

    @passo_reamostragem_cenarios.setter
    def passo_reamostragem_cenarios(self, dado: int):
        b = self.__bloco_por_tipo(BlocoReamostragemCenarios, 0)
        if b is not None:
            b.passo_reamostragem_cenarios = dado

    @property
    def converge_no_zero(self) -> Optional[int]:
        """
        Configuração da linha número 68 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoConvergeNoZero, 0)
        if b is not None:
            return b.valor
        return None

    @converge_no_zero.setter
    def converge_no_zero(self, dado: int):
        b = self.__bloco_por_tipo(BlocoConvergeNoZero, 0)
        if b is not None:
            b.valor = dado

    @property
    def consulta_fcf(self) -> Optional[int]:
        """
        Configuração da linha número 69 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoConsultaFCF, 0)
        if b is not None:
            return b.valor
        return None

    @consulta_fcf.setter
    def consulta_fcf(self, dado: int):
        b = self.__bloco_por_tipo(BlocoConsultaFCF, 0)
        if b is not None:
            b.valor = dado

    @property
    def impressao_ena(self) -> Optional[int]:
        """
        Configuração da linha número 70 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImpressaoENA, 0)
        if b is not None:
            return b.valor
        return None

    @impressao_ena.setter
    def impressao_ena(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImpressaoENA, 0)
        if b is not None:
            b.valor = dado

    @property
    def impressao_cortes_ativos_sim_final(self) -> Optional[int]:
        """
        Configuração da linha número 71 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoImpressaoCortesAtivosSimFinal, 0)
        if b is not None:
            return b.valor
        return None

    @impressao_cortes_ativos_sim_final.setter
    def impressao_cortes_ativos_sim_final(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImpressaoCortesAtivosSimFinal, 0)
        if b is not None:
            b.valor = dado

    @property
    def representacao_agregacao(self) -> Optional[int]:
        """
        Configuração da linha número 72 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoRepresentacaoAgregacao, 0)
        if b is not None:
            return b.valor
        return None

    @representacao_agregacao.setter
    def representacao_agregacao(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRepresentacaoAgregacao, 0)
        if b is not None:
            b.valor = dado

    @property
    def matriz_correlacao_espacial(self) -> Optional[int]:
        """
        Configuração da linha número 73 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoMatrizCorrelacaoEspacial, 0)
        if b is not None:
            return b.valor
        return None

    @matriz_correlacao_espacial.setter
    def matriz_correlacao_espacial(self, dado: int):
        b = self.__bloco_por_tipo(BlocoMatrizCorrelacaoEspacial, 0)
        if b is not None:
            b.valor = dado

    @property
    def desconsidera_convergencia_estatistica(self) -> Optional[int]:
        """
        Configuração da linha número 74 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoDesconsideraConvEstatistica, 0)
        if b is not None:
            return b.valor
        return None

    @desconsidera_convergencia_estatistica.setter
    def desconsidera_convergencia_estatistica(self, dado: int):
        b = self.__bloco_por_tipo(BlocoDesconsideraConvEstatistica, 0)
        if b is not None:
            b.valor = dado

    @property
    def momento_reamostragem(self) -> Optional[int]:
        """
        Configuração da linha número 75 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoMomentoReamostragem, 0)
        if b is not None:
            return b.valor
        return None

    @momento_reamostragem.setter
    def momento_reamostragem(self, dado: int):
        b = self.__bloco_por_tipo(BlocoMomentoReamostragem, 0)
        if b is not None:
            b.valor = dado

    @property
    def mantem_arquivos_energias(self) -> Optional[int]:
        """
        Configuração da linha número 76 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoMantemArquivosEnergias, 0)
        if b is not None:
            return b.valor
        return None

    @mantem_arquivos_energias.setter
    def mantem_arquivos_energias(self, dado: int):
        b = self.__bloco_por_tipo(BlocoMantemArquivosEnergias, 0)
        if b is not None:
            b.valor = dado

    @property
    def inicio_teste_convergencia(self) -> Optional[int]:
        """
        Configuração da linha número 77 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoInicioTesteConvergencia, 0)
        if b is not None:
            return b.valor
        return None

    @inicio_teste_convergencia.setter
    def inicio_teste_convergencia(self, dado: int):
        b = self.__bloco_por_tipo(BlocoInicioTesteConvergencia, 0)
        if b is not None:
            b.valor = dado

    @property
    def sazonaliza_vmint(self) -> Optional[int]:
        """
        Configuração da linha número 78 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSazonalizarVminT, 0)
        if b is not None:
            return b.valor
        return None

    @sazonaliza_vmint.setter
    def sazonaliza_vmint(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSazonalizarVminT, 0)
        if b is not None:
            b.valor = dado

    @property
    def sazonaliza_vmaxt(self) -> Optional[int]:
        """
        Configuração da linha número 79 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSazonalizarVmaxT, 0)
        if b is not None:
            return b.valor
        return None

    @sazonaliza_vmaxt.setter
    def sazonaliza_vmaxt(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSazonalizarVmaxT, 0)
        if b is not None:
            b.valor = dado

    @property
    def sazonaliza_vminp(self) -> Optional[int]:
        """
        Configuração da linha número 80 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSazonalizarVminP, 0)
        if b is not None:
            return b.valor
        return None

    @sazonaliza_vminp.setter
    def sazonaliza_vminp(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSazonalizarVminP, 0)
        if b is not None:
            b.valor = dado

    @property
    def sazonaliza_cfuga_cmont(self) -> Optional[int]:
        """
        Configuração da linha número 81 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoSazonalizarCfugaCmont, 0)
        if b is not None:
            return b.valor
        return None

    @sazonaliza_cfuga_cmont.setter
    def sazonaliza_cfuga_cmont(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSazonalizarCfugaCmont, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricoes_emissao_gee(self) -> Optional[int]:
        """
        Configuração da linha número 82 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoRestricoesEmissaoGEE, 0)
        if b is not None:
            return b.valor
        return None

    @restricoes_emissao_gee.setter
    def restricoes_emissao_gee(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricoesEmissaoGEE, 0)
        if b is not None:
            b.valor = dado

    @property
    def consideracao_media_anual_afluencias(self) -> Optional[int]:
        """
        Configuração da linha número 83 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoAfluenciaAnualPARp, 0)
        if b is not None:
            return b.consideracao_media_anual_afluencias
        return None

    @consideracao_media_anual_afluencias.setter
    def consideracao_media_anual_afluencias(self, dado: int):
        b = self.__bloco_por_tipo(BlocoAfluenciaAnualPARp, 0)
        if b is not None:
            b.consideracao_media_anual_afluencias = dado

    @property
    def reducao_automatica_ordem(self) -> Optional[int]:
        """
        Configuração da linha número 83 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoAfluenciaAnualPARp, 0)
        if b is not None:
            return b.reducao_automatica_ordem
        return None

    @reducao_automatica_ordem.setter
    def reducao_automatica_ordem(self, dado: int):
        b = self.__bloco_por_tipo(BlocoAfluenciaAnualPARp, 0)
        if b is not None:
            b.reducao_automatica_ordem = dado

    @property
    def restricoes_fornecimento_gas(self) -> Optional[int]:
        """
        Configuração da linha número 84 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoRestricoesFornecGas, 0)
        if b is not None:
            return b.valor
        return None

    @restricoes_fornecimento_gas.setter
    def restricoes_fornecimento_gas(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricoesFornecGas, 0)
        if b is not None:
            b.valor = dado

    @property
    def memoria_calculo_cortes(self) -> Optional[int]:
        """
        Configuração da linha número 85 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoMemCalculoCortes, 0)
        if b is not None:
            return b.valor
        return None

    @memoria_calculo_cortes.setter
    def memoria_calculo_cortes(self, dado: int):
        b = self.__bloco_por_tipo(BlocoMemCalculoCortes, 0)
        if b is not None:
            b.valor = dado

    @property
    def considera_geracao_eolica(self) -> Optional[int]:
        """
        Configuração da linha número 86 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGeracaoEolica, 0)
        if b is not None:
            return b.considera
        return None

    @considera_geracao_eolica.setter
    def considera_geracao_eolica(self, dado: int):
        b = self.__bloco_por_tipo(BlocoGeracaoEolica, 0)
        if b is not None:
            b.considera = dado

    @property
    def penalidade_corte_geracao_eolica(self) -> Optional[float]:
        """
        Configuração da linha número 86 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoGeracaoEolica, 0)
        if b is not None:
            return b.penalidade
        return None

    @penalidade_corte_geracao_eolica.setter
    def penalidade_corte_geracao_eolica(self, dado: float):
        b = self.__bloco_por_tipo(BlocoGeracaoEolica, 0)
        if b is not None:
            b.penalidade = dado

    @property
    def compensacao_correlacao_cruzada(self) -> Optional[int]:
        """
        Configuração da linha número 87 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int
        """
        b = self.__bloco_por_tipo(BlocoCompensacaoCorrelacaoCruzada, 0)
        if b is not None:
            return b.valor
        return None

    @compensacao_correlacao_cruzada.setter
    def compensacao_correlacao_cruzada(self, dado: int):
        b = self.__bloco_por_tipo(BlocoCompensacaoCorrelacaoCruzada, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricao_turbinamento(self) -> Optional[int]:
        """
        Configuração da linha número 88 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoConsideracaoTurbinamentoMinimoMaximo, 0)
        if b is not None:
            return b.valor
        return None

    @restricao_turbinamento.setter
    def restricao_turbinamento(self, dado: int):
        b = self.__bloco_por_tipo(BlocoConsideracaoTurbinamentoMinimoMaximo, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricao_defluencia(self) -> Optional[int]:
        """
        Configuração da linha número 89 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoConsideracaoDefluenciaMaxima, 0)
        if b is not None:
            return b.valor
        return None

    @restricao_defluencia.setter
    def restricao_defluencia(self, dado: int):
        b = self.__bloco_por_tipo(BlocoConsideracaoDefluenciaMaxima, 0)
        if b is not None:
            b.valor = dado

    @property
    def aproveitamento_bases_backward(self) -> Optional[int]:
        """
        Configuração da linha número 90 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoAproveitamentoBasePLsBackward, 0)
        if b is not None:
            return b.valor
        return None

    @aproveitamento_bases_backward.setter
    def aproveitamento_bases_backward(self, dado: int):
        b = self.__bloco_por_tipo(BlocoAproveitamentoBasePLsBackward, 0)
        if b is not None:
            b.valor = dado

    @property
    def impressao_estados_geracao_cortes(self) -> Optional[int]:
        """
        Configuração da linha número 91 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoImpressaoEstadosGeracaoCortes, 0)
        if b is not None:
            return b.valor
        return None

    @impressao_estados_geracao_cortes.setter
    def impressao_estados_geracao_cortes(self, dado: int):
        b = self.__bloco_por_tipo(BlocoImpressaoEstadosGeracaoCortes, 0)
        if b is not None:
            b.valor = dado

    @property
    def semente_forward(self) -> Optional[int]:
        """
        Configuração da linha número 92 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoSementeForward, 0)
        if b is not None:
            return b.valor
        return None

    @semente_forward.setter
    def semente_forward(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSementeForward, 0)
        if b is not None:
            b.valor = dado

    @property
    def semente_backward(self) -> Optional[int]:
        """
        Configuração da linha número 93 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoSementeBackward, 0)
        if b is not None:
            return b.valor
        return None

    @semente_backward.setter
    def semente_backward(self, dado: int):
        b = self.__bloco_por_tipo(BlocoSementeBackward, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricao_lpp_turbinamento_maximo_ree(self) -> Optional[int]:
        """
        Configuração da linha número 94 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoRestricaoLPPTurbinamentoMaximoREE, 0)
        if b is not None:
            return b.valor
        return None

    @restricao_lpp_turbinamento_maximo_ree.setter
    def restricao_lpp_turbinamento_maximo_ree(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricaoLPPTurbinamentoMaximoREE, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricao_lpp_defluencia_maxima_ree(self) -> Optional[int]:
        """
        Configuração da linha número 95 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoRestricaoLPPDefluenciaMaximaREE, 0)
        if b is not None:
            return b.valor
        return None

    @restricao_lpp_defluencia_maxima_ree.setter
    def restricao_lpp_defluencia_maxima_ree(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricaoLPPDefluenciaMaximaREE, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricao_lpp_turbinamento_maximo_uhe(self) -> Optional[int]:
        """
        Configuração da linha número 96 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoRestricaoLPPTurbinamentoMaximoUHE, 0)
        if b is not None:
            return b.valor
        return None

    @restricao_lpp_turbinamento_maximo_uhe.setter
    def restricao_lpp_turbinamento_maximo_uhe(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricaoLPPTurbinamentoMaximoUHE, 0)
        if b is not None:
            b.valor = dado

    @property
    def restricao_lpp_defluencia_maxima_uhe(self) -> Optional[int]:
        """
        Configuração da linha número 97 do arquivo `dger.dat`.

        :return: O valor do campo
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoRestricaoLPPDefluenciaMaximaUHE, 0)
        if b is not None:
            return b.valor
        return None

    @restricao_lpp_defluencia_maxima_uhe.setter
    def restricao_lpp_defluencia_maxima_uhe(self, dado: int):
        b = self.__bloco_por_tipo(BlocoRestricaoLPPDefluenciaMaximaUHE, 0)
        if b is not None:
            b.valor = dado



# Rotinas de testes associadas ao arquivo MEDIAS-MERC.CSV do NWLISTOP
from inewave.nwlistop.mediasmerc import MediasMerc
import pandas as pd  # type: ignore


leitor = MediasMerc.le_arquivo("tests/_arquivos")


def test_eq_mediasmerc():
    leitor2 = MediasMerc.le_arquivo("tests/_arquivos")
    assert leitor == leitor2


def test_neq_mediasmerc():
    leitor2 = MediasMerc.le_arquivo("tests/_arquivos")
    leitor2.medias = pd.DataFrame()
    assert leitor != leitor2



MockRegistroEolicaConfiguracaoPeriodo = (
    "EOLICA-CONFIGURACAO-PERIODO ; 1 ; 2021/01; 2030/12; centralizado\n"
)


MockEolicaConfig = [
    "&EOLICA-CONFIGURACAO-PERIODO; CodigoEolica; PeriodoInicial; PeriodoFinal  ; EstadoEolica (centralizado = 1, fixo = 2, nãoexistente = 3)\n",
    "EOLICA-CONFIGURACAO-PERIODO ; 1 ; 2021/01; 2030/12; centralizado\n",
    "EOLICA-CONFIGURACAO-PERIODO ; 2 ; 2021/01; 2030/12; centralizado\n",
    "EOLICA-CONFIGURACAO-PERIODO ; 3 ; 2021/01; 2030/12; centralizado\n",
    "EOLICA-CONFIGURACAO-PERIODO ; 4 ; 2021/01; 2030/12; centralizado\n",
    "EOLICA-CONFIGURACAO-PERIODO ; 5 ; 2021/01; 2030/12; centralizado\n",
]



from inewave.nwlistop.modelos.eafbsin import EafsAnos

from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class EafbSIN(ArquivoSIN):
    """
    Armazena os dados das saídas referentes à energia natural
    afluente para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `eafbsin.out`.
    """

    BLOCKS = [
        EafsAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="eafbsin.out"
    ) -> "EafbSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="eafbsin.out"):
        self.write(diretorio, nome_arquivo)



from inewave.nwlistop.modelos.blocos.parsubmercados import ParSubmercados
from inewave.nwlistop.modelos.arquivos.arquivoparsubmercadopatamar import (
    ArquivoParSubmercadoPatamar,
)
from inewave.nwlistop.modelos.intercambio import IntercambioAnos


class Intercambio(ArquivoParSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes ao intercâmbio
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `int00x00y.out`, onde x e y variam
    conforme os submercados em questão.

    """

    BLOCKS = [
        ParSubmercados,
        IntercambioAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="int001002.out"
    ) -> "Intercambio":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="ghtotm001.out"):
        self.write(diretorio, nome_arquivo)



from inewave.newave.modelos.parp import (
    BlocoSerieEnergiaREE,
    BlocoCorrelEnergiasREE,
    BlocoCorrelParcialEnergiasREE,
    BlocoOrdemModeloREE,
    BlocoCoeficientesModeloREE,
    BlocoSerieRuidosREE,
    BlocoCorrelRuidosREE,
    BlocoSerieMediasREE,
    BlocoCorrelCruzadaMediaREE,
    BlocoCorrelEspacialAnualConfig,
    BlocoCorrelEspacialMensalConfig,
)

from inewave.newave import PARp


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.parp import (
    MockSeriesEnergiaPARp,
    MockCorrelacaoSeriesEnergiasREE,
    MockCorrelacaoParcialSeriesEnergiaREE,
    MockOrdemOriginalModeloREE,
    MockCoeficientesModeloREE,
    MockSerieRuidosREE,
    MockCorrelSerieRuidosREE,
    MockSerieMedia12MesesREE,
    MockCorrelCruzadaMedia12Meses,
    MockCorrelEspacialAnual,
    MockCorrelEspacialMensal,
    MockPARp,
)


def test_series_energia_ree():

    m: MagicMock = mock_open(read_data="".join(MockSeriesEnergiaPARp))
    b = BlocoSerieEnergiaREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 87
    assert b.data.shape[1] == 15
    assert b.data.iloc[0, 3] == 8536.82
    assert b.data.iloc[-1, -1] == 5616.44


def test_correlacao_series_energia_ree():

    m: MagicMock = mock_open(
        read_data="".join(MockCorrelacaoSeriesEnergiasREE)
    )
    b = BlocoCorrelEnergiasREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 120
    assert b.data.shape[1] == 12
    assert b.data.iloc[0, 1] == 0.28570
    assert b.data.iloc[-1, -1] == 0.04142


def test_correlacao_parcial_series_energia_ree():

    m: MagicMock = mock_open(
        read_data="".join(MockCorrelacaoParcialSeriesEnergiaREE)
    )
    b = BlocoCorrelParcialEnergiasREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 120
    assert b.data.shape[1] == 12
    assert b.data.iloc[0, 1] == 0.19220
    assert b.data.iloc[-1, -1] == -0.15513


def test_ordem_modelo_ree():

    m: MagicMock = mock_open(read_data="".join(MockOrdemOriginalModeloREE))
    b = BlocoOrdemModeloREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 10
    assert b.data.shape[1] == 14
    assert b.data.iloc[0, 2] == 1
    assert b.data.iloc[-1, -1] == 4


def test_coeficientes_modelo_ree():

    m: MagicMock = mock_open(read_data="".join(MockCoeficientesModeloREE))
    b = BlocoCoeficientesModeloREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 1
    assert b.data.shape[1] == 24
    assert b.data.loc[0, "Psi 1"] == 0.207
    assert b.data.loc[0, "Psi Norm 1"] == 0.263
    assert b.data.loc[0, "Psi A"] == 0.177
    assert b.data.loc[0, "Psi Norm A"] == 0.485


def test_serie_ruidos_ree():

    m: MagicMock = mock_open(read_data="".join(MockSerieRuidosREE))
    b = BlocoSerieRuidosREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 87
    assert b.data.shape[1] == 14
    assert b.data.iloc[0, 1] == 0.0
    assert b.data.iloc[-1, -1] == -0.152


def test_correl_serie_ruidos_ree():

    m: MagicMock = mock_open(read_data="".join(MockCorrelSerieRuidosREE))
    b = BlocoCorrelRuidosREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 120
    assert b.data.shape[1] == 12
    assert b.data.iloc[0, 1] == -0.05921
    assert b.data.iloc[-1, -1] == -0.08159


def test_serie_media_12_meses_ree():

    m: MagicMock = mock_open(read_data="".join(MockSerieMedia12MesesREE))
    b = BlocoSerieMediasREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 87
    assert b.data.shape[1] == 14
    assert b.data.iloc[0, 1] == 0.0000
    assert b.data.iloc[-1, -1] == 3136.70


def test_correl_cruzada_media_12_meses_ree():

    m: MagicMock = mock_open(read_data="".join(MockCorrelCruzadaMedia12Meses))
    b = BlocoCorrelCruzadaMediaREE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 120
    assert b.data.shape[1] == 13
    assert b.data.iloc[0, 1] == 0.26913
    assert b.data.iloc[-1, -1] == 0.76845


def test_correl_espacial_anual():

    m: MagicMock = mock_open(read_data="".join(MockCorrelEspacialAnual))
    b = BlocoCorrelEspacialAnualConfig()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 12
    assert b.data.shape[1] == 14
    assert b.data.iloc[0, 2] == 1.000
    assert b.data.iloc[-1, -1] == 1.000


def test_correl_espacial_mensal():

    m: MagicMock = mock_open(read_data="".join(MockCorrelEspacialMensal))
    b = BlocoCorrelEspacialMensalConfig()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 12
    assert b.data.shape[1] == 15
    assert b.data.iloc[0, 3] == 1.0000
    assert b.data.iloc[-1, -1] == 0.0671


def test_atributos_encontrados_parp():
    m: MagicMock = mock_open(read_data="".join(MockPARp))
    with patch("builtins.open", m):
        parp = PARp.le_arquivo("")
        assert parp.series_energia_ree is not None
        assert parp.correlacao_series_energia_ree is not None
        assert parp.correlacao_parcial_series_energia_ree is not None
        assert parp.ordem_original_modelo is not None
        assert parp.ordem_final_modelo is not None
        assert parp.coeficientes is not None
        assert parp.series_media_ree is not None
        assert parp.series_ruido_ree is not None
        assert parp.correlacao_series_ruidos_ree is not None
        assert parp.correlacao_cruzada_media_ree is not None
        assert parp.correlacao_espacial_anual is not None
        assert parp.correlacao_espacial_mensal is not None


def test_atributos_nao_encontrados_parp():
    m: MagicMock = mock_open(read_data="".join(MockSeriesEnergiaPARp))
    with patch("builtins.open", m):
        parp = PARp.le_arquivo("")
        assert parp.series_energia_ree is not None
        assert parp.correlacao_series_energia_ree is None
        assert parp.correlacao_parcial_series_energia_ree is None
        assert parp.ordem_original_modelo is None
        assert parp.ordem_final_modelo is None
        assert parp.coeficientes is None
        assert parp.series_media_ree is None
        assert parp.series_ruido_ree is None
        assert parp.correlacao_series_ruidos_ree is None
        assert parp.correlacao_cruzada_media_ree is None
        assert parp.correlacao_espacial_anual is None
        assert parp.correlacao_espacial_mensal is None


def test_eq_parp():
    m: MagicMock = mock_open(read_data="".join(MockPARp))
    with patch("builtins.open", m):
        parp1 = PARp.le_arquivo("")
        parp2 = PARp.le_arquivo("")
        assert parp1 == parp2


def test_neq_parp():
    m: MagicMock = mock_open(read_data="".join(MockPARp))
    with patch("builtins.open", m):
        parp1 = PARp.le_arquivo("")
        parp2 = PARp.le_arquivo("")
        parp2.series_energia_ree.iloc[0, 0] = -1
        assert parp1 != parp2



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type


class GEE(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações
    de emissão de Gases de Efeito Estufa (GEE).

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = []

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="gee.dat") -> "GEE":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="gee.dat"):
        self.write(diretorio, nome_arquivo)



MockArquivos = [
    "DADOS GERAIS                : dger.dat\n",
    "DADOS DOS SUBSISTEMAS       : sistema.dat\n",
    "CONFIGURACAO HIDRAULICA     : confhd.dat\n",
    "ALTERACAO DADOS USINAS HIDRO: modif.dat\n",
    "CONFIGURACAO TERMICA        : conft.dat\n",
    "DADOS DAS USINAS TERMICAS   : term.dat\n",
    "DADOS DAS CLASSES TERMICAS  : clast.dat\n",
    "DADOS DE EXPANSAO HIDRAULICA: exph.dat\n",
    "ARQUIVO DE EXPANSAO TERMICA : expt.dat\n",
    "ARQUIVO DE PATAMARES MERCADO: patamar.dat\n",
    "ARQUIVO DE CORTES DE BENDERS: cortes.dat\n",
    "ARQUIVO DE CABECALHO CORTES : cortesh.dat\n",
    "RELATORIO DE CONVERGENCIA   : pmo.dat\n",
    "RELATORIO DE E. SINTETICAS  : parp.dat\n",
    "RELATORIO DETALHADO FORWARD : forward.dat\n",
    "ARQUIVO DE CABECALHO FORWARD: forwarh.dat\n",
    "ARQUIVO DE S.HISTORICAS S.F.: shist.dat\n",
    "ARQUIVO DE MANUT.PROG. UTE'S: manutt.dat\n",
    "ARQUIVO P/DESPACHO HIDROTERM: newdesp.dat\n",
    "ARQUIVO C/TEND. HIDROLOGICA : vazpast.dat\n",
    "ARQUIVO C/DADOS DE ITAIPU   : itaipu.dat\n",
    "ARQUIVO C/DEMAND S. BIDDING : bid.dat\n",
    "ARQUIVO C/CARGAS ADICIONAIS : c_adic.dat\n",
    "ARQUIVO C/FATORES DE PERDAS : loss.dat\n",
    "ARQUIVO C/PATAMARES GTMIN   : gtminpat.dat\n",
    "ARQUIVO ENSO 1              : elnino.dat\n",
    "ARQUIVO ENSO 2              : ensoaux.dat\n",
    "ARQUIVO DSVAGUA             : dsvagua.dat\n",
    "ARQUIVO P/PENALID. POR DESV.: penalid.dat\n",
    "ARQUIVO C.GUIA / PENAL.VMINT: curva.dat\n",
    "ARQUIVO AGRUPAMENTO LIVRE   : agrint.dat\n",
    "ARQUIVO DESP. ANTEC. GNL    : adterm.dat\n",
    "ARQUIVO GER. HIDR. MIN      : ghmin.dat\n",
    "ARQUIVO AVERSAO RISCO - SAR : sar.dat\n",
    "ARQUIVO AVERSAO RISCO - CVAR: cvar.dat\n",
    "DADOS DOS RESER.EQ.ENERGIA  : ree.dat\n",
    "ARQUIVO RESTRICOES ELETRICAS: re.dat\n",
    "ARQUIVO DE TECNOLOGIAS      : tecno.dat\n",
    "DADOS DE ABERTURAS          : abertura.dat\n",
    "ARQUIVO DE EMISSOES GEE     : gee.dat\n",
    "ARQUIVO DE RESTRICAO DE GAS : clasgas.dat\n",
]



from inewave.config import (
    MAX_ANOS_ESTUDO,
    MAX_PATAMARES,
    MAX_SUBMERCADOS,
    MESES_DF,
)

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoNumeroPatamares(Section):
    """
    Bloco com o número de patamares de carga considerados
    no caso.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(2, 1),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoNumeroPatamares):
            return False
        bloco: BlocoNumeroPatamares = o
        if not all(
            [
                isinstance(self.data, int),
                isinstance(o.data, int),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.data = self.__linha.read(file.readline())[0]

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, int):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )
        file.write(self.__linha.write([self.data]))


class BlocoDuracaoPatamar(Section):
    """
    Bloco com a duração de cada patamar por mês
    de estudo, extraído do arquivo `patamar.dat`.
    """

    FIM_BLOCO = "SUBSISTEMA"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_ano: List[Field] = [IntegerField(4, 0)]
        campos_duracao: List[Field] = [
            FloatField(6, 6 + i * 8, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_duracao)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoDuracaoPatamar):
            return False
        bloco: BlocoDuracaoPatamar = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=["Ano"] + MESES_DF)
            df = df.astype({"Ano": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        ano_atual = 0
        tabela = np.zeros(
            (MAX_SUBMERCADOS * MAX_ANOS_ESTUDO, len(MESES_DF) + 1)
        )
        while True:
            ultima_linha = file.tell()
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoDuracaoPatamar.FIM_BLOCO in linha:
                file.seek(ultima_linha)
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                tabela[i, 0] = ano_atual
                tabela[i, 1:] = dados[1:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        ultimo_ano = 0
        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            ano_linha = (
                int(linha_lida["Ano"])
                if linha_lida["Ano"] != ultimo_ano
                else None
            )
            ultimo_ano = int(linha_lida["Ano"])
            file.write(
                self.__linha.write([ano_linha] + linha_lida[MESES_DF].tolist())
            )


class BlocoCargaPatamar(Section):
    """
    Bloco com a carga de cada patamar por mês
    de estudo, extraído do arquivo `patamar.dat`.
    """

    FIM_BLOCO = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(4, 0)])
        campo_ano: List[Field] = [IntegerField(4, 3)]
        campo_carga: List[Field] = [
            FloatField(6, 8 + i * 7, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campo_carga)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCargaPatamar):
            return False
        bloco: BlocoCargaPatamar = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=["Subsistema", "Ano"] + MESES_DF)
            df = df.astype({"Subsistema": "int64", "Ano": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(4):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_atual = 0
        ano_atual = 0
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_ANOS_ESTUDO * MAX_PATAMARES,
                len(MESES_DF) + 2,
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoCargaPatamar.FIM_BLOCO in linha[:4]:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 8:
                subsis_atual = self.__linha_subsis.read(linha)[0]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                tabela[i, 0] = subsis_atual
                tabela[i, 1] = ano_atual
                tabela[i, 2:] = dados[1:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        ultimo_subsistema = 0
        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if linha_lida["Subsistema"] != ultimo_subsistema:
                ultimo_subsistema = linha_lida["Subsistema"]
                ultimo_ano = 0
                file.write(self.__linha_subsis.write([int(ultimo_subsistema)]))
            ano_linha = (
                int(linha_lida["Ano"])
                if linha_lida["Ano"] != ultimo_ano
                else None
            )
            ultimo_ano = int(linha_lida["Ano"])
            file.write(
                self.__linha.write([ano_linha] + linha_lida[MESES_DF].tolist())
            )
        file.write(BlocoCargaPatamar.FIM_BLOCO + "\n")


class BlocoIntercambioPatamarSubsistemas(Section):
    """
    Bloco com o fator de correção do intercâmbio entre subsistemas para
    cada patamar, por mês de estudo, extraído do arquivo `patamar.dat`.
    """

    FIM_BLOCO = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(3, 1), IntegerField(3, 5)])
        campo_ano: List[Field] = [IntegerField(4, 3)]
        campos_intercambios: List[Field] = [
            FloatField(6, 8 + i * 7, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_intercambios)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoIntercambioPatamarSubsistemas):
            return False
        bloco: BlocoIntercambioPatamarSubsistemas = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                tabela,
                columns=["Subsistema De", "Subsistema Para", "Ano"] + MESES_DF,
            )
            df = df.astype(
                {
                    "Subsistema De": "int64",
                    "Subsistema Para": "int64",
                    "Ano": "int64",
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(5):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_de_atual = 0
        subsis_para_atual = 0
        ano_atual = 0
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_SUBMERCADOS * MAX_ANOS_ESTUDO,
                len(MESES_DF) + 3,
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if (
                len(linha) < 3
                or BlocoIntercambioPatamarSubsistemas.FIM_BLOCO in linha[:4]
            ):
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 12:
                dados = self.__linha_subsis.read(linha)
                subsis_de_atual = dados[0]
                subsis_para_atual = dados[1]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                tabela[i, 0] = subsis_de_atual
                tabela[i, 1] = subsis_para_atual
                tabela[i, 2] = ano_atual
                tabela[i, 3:] = dados[1:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        ultimo_subsistema_de = 0
        ultimo_subsistema_para = 0

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if any(
                [
                    linha_lida["Subsistema De"] != ultimo_subsistema_de,
                    linha_lida["Subsistema Para"] != ultimo_subsistema_para,
                ]
            ):
                ultimo_ano = 0
                ultimo_subsistema_de = linha_lida["Subsistema De"]
                ultimo_subsistema_para = linha_lida["Subsistema Para"]
                file.write(
                    self.__linha_subsis.write(
                        [
                            int(ultimo_subsistema_de),
                            int(ultimo_subsistema_para),
                        ]
                    )
                )
            ano_linha = (
                int(linha_lida["Ano"])
                if linha_lida["Ano"] != ultimo_ano
                else None
            )
            ultimo_ano = int(linha_lida["Ano"])
            file.write(
                self.__linha.write([ano_linha] + linha_lida[MESES_DF].tolist())
            )
        file.write(BlocoIntercambioPatamarSubsistemas.FIM_BLOCO + "\n")


class BlocoUsinasNaoSimuladas(Section):
    """
    Bloco com a geração de usinas não simuladas em P.U. para
    cada patamar, por mês de estudo, extraído do arquivo `patamar.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(3, 1), IntegerField(3, 5)])
        campo_ano: List[Field] = [IntegerField(4, 3)]
        campos_usinas: List[Field] = [
            FloatField(6, 8 + i * 7, 4) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_usinas)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUsinasNaoSimuladas):
            return False
        bloco: BlocoUsinasNaoSimuladas = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                tabela, columns=["Subsistema", "Bloco", "Ano"] + MESES_DF
            )
            df = df.astype(
                {"Subsistema": "int64", "Bloco": "int64", "Ano": "int64"}
            )
            return df

        # Salta as linhas adicionais
        for _ in range(4):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_atual = 0
        bloco_atual = 0
        ano_atual = 0
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_SUBMERCADOS * MAX_ANOS_ESTUDO,
                len(MESES_DF) + 3,
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 12:
                dados = self.__linha_subsis.read(linha)
                subsis_atual = dados[0]
                bloco_atual = dados[1]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                tabela[i, 0] = subsis_atual
                tabela[i, 1] = bloco_atual
                tabela[i, 2] = ano_atual
                tabela[i, 3:] = dados[1:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do patamar.dat não foram lidos com sucesso"
            )

        ultimo_subsistema = 0
        ultimo_bloco = 0

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if any(
                [
                    linha_lida["Subsistema"] != ultimo_subsistema,
                    linha_lida["Bloco"] != ultimo_bloco,
                ]
            ):
                ultimo_ano = 0
                ultimo_subsistema = linha_lida["Subsistema"]
                ultimo_bloco = linha_lida["Bloco"]
                file.write(
                    self.__linha_subsis.write(
                        [
                            int(ultimo_subsistema),
                            int(ultimo_bloco),
                        ]
                    )
                )
            ano_linha = (
                int(linha_lida["Ano"])
                if linha_lida["Ano"] != ultimo_ano
                else None
            )
            ultimo_ano = int(linha_lida["Ano"])
            file.write(
                self.__linha.write([ano_linha] + linha_lida[MESES_DF].tolist())
            )
        file.write(BlocoIntercambioPatamarSubsistemas.FIM_BLOCO + "\n")



from inewave.newave.modelos.re import (
    BlocoUsinasConjuntoRE,
    BlocoConfiguracaoRestricoesRE,
)

from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class RE(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às restrições
    elétricas existentes.

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoUsinasConjuntoRE,
        BlocoConfiguracaoRestricoesRE,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="re.dat") -> "RE":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="re.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def usinas_conjuntos(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os conjuntos de usinas com restrições elétricas.

        - Conjunto (`int`)
        - Usina 1 (`float`)
        - Usina 2 (`float`)
        - ...
        - Usina N (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoUsinasConjuntoRE, 0)
        if b is not None:
            return b.data
        return None

    @usinas_conjuntos.setter
    def usinas_conjuntos(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoUsinasConjuntoRE, 0)
        if b is not None:
            b.data = df

    @property
    def restricoes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as configurações das restrições elétricas.

        - Conjunto (`int`)
        - Mês Início (`int`)
        - Mês Fim (`int`)
        - Ano Início (`int`)
        - Ano Fim (`int`)
        - Flag P (`int`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoConfiguracaoRestricoesRE, 0)
        if b is not None:
            return b.data
        return None

    @restricoes.setter
    def restricoes(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoConfiguracaoRestricoesRE, 0)
        if b is not None:
            b.data = df



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.datetimefield import DatetimeField
from cfinterface.components.floatfield import FloatField
from datetime import datetime


class RegistroEolicaHistoricoVentoHorizonte(Register):
    """
    Registro que contém o horizonte de históricos de vento.
    """

    IDENTIFIER = "EOLICA-HISTORICO-VENTO-HORIZONTE"
    IDENTIFIER_DIGITS = 32
    LINE = Line(
        [
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
        ],
        delimiter=";",
    )

    @property
    def data_inicial(self) -> Optional[datetime]:
        return self.data[0]

    @data_inicial.setter
    def data_inicial(self, c: datetime):
        self.data[0] = c

    @property
    def data_final(self) -> Optional[datetime]:
        return self.data[1]

    @data_final.setter
    def data_final(self, c: datetime):
        self.data[1] = c


class RegistroEolicaHistoricoVento(Register):
    """
    Registro que contém os históricos de vento para o horizonte
    considerado.
    """

    IDENTIFIER = "EOLICA-HISTORICO-VENTO"
    IDENTIFIER_DIGITS = 22
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(decimal_digits=2),
            FloatField(decimal_digits=2),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def data_inicial(self) -> Optional[datetime]:
        return self.data[1]

    @data_inicial.setter
    def data_inicial(self, c: datetime):
        self.data[1] = c

    @property
    def data_final(self) -> Optional[datetime]:
        return self.data[2]

    @data_final.setter
    def data_final(self, c: datetime):
        self.data[2] = c

    @property
    def velocidade(self) -> Optional[float]:
        return self.data[3]

    @velocidade.setter
    def velocidade(self, c: float):
        self.data[3] = c

    @property
    def direcao(self) -> Optional[float]:
        return self.data[4]

    @direcao.setter
    def direcao(self, c: float):
        self.data[4] = c



# Imports do próprio módulo
from inewave._utils.leituracsv import LeituraCSV

# Imports de módulos externos


class LeituraMediasMerc(LeituraCSV):
    """
    Realiza a leitura do arquivo MEDIAS-MERC.CSV
    existente em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos do arquivo MEDIAS-MERC.CSV, construindo um
    objeto `MediasMerc` cujas informações são as mesmas do arquivo.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    """

    def __init__(self, diretorio: str) -> None:
        super().__init__(diretorio)

    def processa_dados_lidos(self):
        # Remove os espaços dos índices
        vars_atuais = list(self._dados.index)
        vars_novas = [v.strip() for v in vars_atuais]
        self._dados.index = vars_novas
        # Renomeia as colunas
        cols_atuais = list(self._dados.columns)
        cols_novas = [c.strip() for c in cols_atuais]
        cols_novas[0] = "Submercado"
        self._dados.columns = cols_novas
        # Exclui a última coluna (em branco)
        self._dados.drop(columns=[cols_novas[-1]], inplace=True)
        # Substitui os elementos da primeira coluna pelos submercados
        mapa_termos = {0: "SIN", 1: "SE", 2: "S", 3: "NE", 4: "N"}
        nova_coluna = [mapa_termos[i] for i in self._dados["Submercado"]]
        self._dados["Submercado"] = nova_coluna



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField


from inewave.nwlistop.modelos.blocos.valoresserie import ValoresSerie


class EarmsAnos(ValoresSerie):
    """
    Bloco com as informações das tabelas de energias armazenada final
    por série e por mês/ano de estudo.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [IntegerField(4, 1)]  # type: ignore
        + [
            FloatField(8, 7 + 10 * i, 1) for i in range(len(MESES_DF) + 1)  # type: ignore
        ]
    )



# Rotinas de testes associadas ao arquivo patamar.dat do NEWAVE
from inewave.newave.modelos.patamar import (
    BlocoNumeroPatamares,
    BlocoDuracaoPatamar,
    BlocoCargaPatamar,
    BlocoIntercambioPatamarSubsistemas,
    BlocoUsinasNaoSimuladas,
)

from inewave.newave import Patamar


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.patamar import (
    MockBlocoNumeroPatamares,
    MockBlocoDuracaoMensalPatamares,
    MockBlocoCargaSubsistema,
    MockBlocoIntercambioSubsistemas,
    MockBlocoUsinasNaoSimuladas,
    MockPatamar,
)


def test_bloco_numero_patamares_patamar():

    m: MagicMock = mock_open(read_data="".join(MockBlocoNumeroPatamares))
    b = BlocoNumeroPatamares()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == 3


def test_bloco_duracao_patamares_patamar():

    m: MagicMock = mock_open(
        read_data="".join(MockBlocoDuracaoMensalPatamares)
    )
    b = BlocoDuracaoPatamar()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1995
    assert b.data.iloc[0, 1] == 0.2366
    assert b.data.iloc[0, 2] == 0.2184
    assert b.data.iloc[0, 3] == 0.2366
    assert b.data.iloc[0, 4] == 0.2778
    assert b.data.iloc[0, 5] == 0.3226
    assert b.data.iloc[0, 6] == 0.3500
    assert b.data.iloc[0, 7] == 0.3710
    assert b.data.iloc[0, 8] == 0.3387
    assert b.data.iloc[0, 9] == 0.2917
    assert b.data.iloc[0, 10] == 0.2823
    assert b.data.iloc[0, 11] == 0.2222
    assert b.data.iloc[0, 12] == 0.2366
    assert b.data.iloc[-1, -1] == 0.5081


def test_bloco_carga_subsistema_patamar():

    m: MagicMock = mock_open(read_data="".join(MockBlocoCargaSubsistema))
    b = BlocoCargaPatamar()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[0, 1] == 1995
    assert b.data.iloc[0, 2] == 1.1492
    assert b.data.iloc[0, 3] == 1.1557
    assert b.data.iloc[0, 4] == 1.1522
    assert b.data.iloc[0, 5] == 1.1612
    assert b.data.iloc[0, 6] == 1.1666
    assert b.data.iloc[0, 7] == 1.1612
    assert b.data.iloc[0, 8] == 1.1530
    assert b.data.iloc[0, 9] == 1.1632
    assert b.data.iloc[0, 10] == 1.1598
    assert b.data.iloc[0, 11] == 1.1531
    assert b.data.iloc[0, 12] == 1.1623
    assert b.data.iloc[0, 13] == 1.1494
    assert b.data.iloc[-1, -1] == 0.9673


def test_bloco_intercambio_patamar():

    m: MagicMock = mock_open(
        read_data="".join(MockBlocoIntercambioSubsistemas)
    )
    b = BlocoIntercambioPatamarSubsistemas()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[0, 1] == 2
    assert b.data.iloc[0, 2] == 1995
    assert b.data.iloc[0, 3] == 1.0000
    assert b.data.iloc[0, 4] == 1.0000
    assert b.data.iloc[0, 5] == 1.0000
    assert b.data.iloc[0, 6] == 1.0000
    assert b.data.iloc[0, 7] == 1.0000
    assert b.data.iloc[0, 8] == 1.0000
    assert b.data.iloc[0, 9] == 0.9720
    assert b.data.iloc[0, 10] == 0.9698
    assert b.data.iloc[0, 11] == 0.9688
    assert b.data.iloc[0, 12] == 0.9681
    assert b.data.iloc[0, 13] == 0.9669
    assert b.data.iloc[0, 14] == 0.9683
    assert b.data.iloc[-1, -1] == 1.0


def test_bloco_usinas_nao_simuladas_patamar():

    m: MagicMock = mock_open(read_data="".join(MockBlocoUsinasNaoSimuladas))
    b = BlocoUsinasNaoSimuladas()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[0, 1] == 1
    assert b.data.iloc[0, 2] == 1995
    assert b.data.iloc[0, 3] == 1.0000
    assert b.data.iloc[0, 4] == 1.0000
    assert b.data.iloc[0, 5] == 1.0000
    assert b.data.iloc[0, 6] == 1.0000
    assert b.data.iloc[0, 7] == 1.0118
    assert b.data.iloc[0, 8] == 1.0152
    assert b.data.iloc[0, 9] == 1.0206
    assert b.data.iloc[0, 10] == 1.0222
    assert b.data.iloc[0, 11] == 1.0150
    assert b.data.iloc[0, 12] == 1.0108
    assert b.data.iloc[0, 13] == 1.0020
    assert b.data.iloc[0, 14] == 0.9965
    assert b.data.iloc[-1, -1] == 1.0


def test_atributos_encontrados_patamar():
    m: MagicMock = mock_open(read_data="".join(MockPatamar))
    with patch("builtins.open", m):
        ad = Patamar.le_arquivo("")
        assert ad.numero_patamares is not None
        assert ad.duracao_mensal_patamares is not None
        assert ad.carga_patamares is not None
        assert ad.intercambio_patamares is not None
        assert ad.usinas_nao_simuladas is not None


def test_atributos_nao_encontrados_patamar():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Patamar.le_arquivo("")
        assert ad.numero_patamares is None
        assert ad.duracao_mensal_patamares is None
        assert ad.carga_patamares is None
        assert ad.intercambio_patamares is None
        assert ad.usinas_nao_simuladas is None


def test_eq_patamar():
    m: MagicMock = mock_open(read_data="".join(MockPatamar))
    with patch("builtins.open", m):
        cf1 = Patamar.le_arquivo("")
        cf2 = Patamar.le_arquivo("")
        assert cf1 == cf2


def test_neq_patamar():
    m: MagicMock = mock_open(read_data="".join(MockPatamar))
    with patch("builtins.open", m):
        cf1 = Patamar.le_arquivo("")
        cf2 = Patamar.le_arquivo("")
        cf2.numero_patamares = 0
        assert cf1 != cf2


def test_leitura_escrita_patamar():
    m_leitura: MagicMock = mock_open(read_data="".join(MockPatamar))
    with patch("builtins.open", m_leitura):
        cf1 = Patamar.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Patamar.le_arquivo("")
        assert cf1 == cf2



# Rotinas de testes associadas ao arquivo eolica-geracao.csv do NEWAVE
from datetime import datetime
from inewave.newave.modelos.eolicageracao import (
    RegistroEolicaGeracaoPatamar,
)

from inewave.newave.eolicageracao import EolicaGeracao

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.eolicageracao import (
    MockRegistroEolicaGeracaoPatamar,
    MockEolicaGeracao,
)


def test_registro_eolica_geracao_patamar_eolicageracao():

    m: MagicMock = mock_open(
        read_data="".join(MockRegistroEolicaGeracaoPatamar)
    )
    r = RegistroEolicaGeracaoPatamar()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(2021, 1, 1), datetime(2021, 1, 1), 2, 1.0496]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 2
    assert r.data_inicial == datetime(2021, 1, 1)
    r.data_inicial = datetime(2022, 1, 1)
    assert r.data_final == datetime(2021, 1, 1)
    r.data_final = datetime(2022, 1, 1)
    assert r.indice_patamar == 2
    r.indice_patamar = 1
    assert r.profundidade == 1.0496
    r.profundidade = 2.0


def test_atributos_encontrados_eolicageracao():
    m: MagicMock = mock_open(read_data="".join(MockEolicaGeracao))
    with patch("builtins.open", m):
        e = EolicaGeracao.le_arquivo("")
        assert len(e.eolica_geracao_profundidade_periodo_patamar()) > 0


def test_eq_eolicageracao():
    m: MagicMock = mock_open(read_data="".join(MockEolicaGeracao))
    with patch("builtins.open", m):
        cf1 = EolicaGeracao.le_arquivo("")
        cf2 = EolicaGeracao.le_arquivo("")
        assert cf1 == cf2


def test_neq_eolicageracao():
    m: MagicMock = mock_open(read_data="".join(MockEolicaGeracao))
    with patch("builtins.open", m):
        cf1 = EolicaGeracao.le_arquivo("")
        cf2 = EolicaGeracao.le_arquivo("")
        cf2.deleta_registro(
            cf1.eolica_geracao_profundidade_periodo_patamar()[0]
        )
        assert cf1 != cf2


def test_leitura_escrita_eolicageracao():
    m_leitura: MagicMock = mock_open(read_data="".join(MockEolicaGeracao))
    with patch("builtins.open", m_leitura):
        cf1 = EolicaGeracao.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = EolicaGeracao.le_arquivo("")
        assert cf1 == cf2



from inewave.nwlistop.intercambio import Intercambio

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.intercambio import MockIntercambio


def test_atributos_encontrados_intercambio():
    m: MagicMock = mock_open(read_data="".join(MockIntercambio))
    with patch("builtins.open", m):
        n = Intercambio.le_arquivo("")
        assert n.submercado_de is not None
        assert n.submercado_de == "SUDESTE"
        assert n.submercado_para == "SUL"
        assert n.valores is not None
        assert n.valores.iloc[0, 0] == 2020
        assert n.valores.iloc[-1, -2] == 2835.2


def test_atributos_nao_encontrados_intercambio():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = Intercambio.le_arquivo("")
        assert n.submercado_de is None
        assert n.submercado_para is None
        assert n.valores is None


def test_eq_intercambio():
    m: MagicMock = mock_open(read_data="".join(MockIntercambio))
    with patch("builtins.open", m):
        n1 = Intercambio.le_arquivo("")
        n2 = Intercambio.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from inewave.config import MAX_RES

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoUsinasConjuntoRE(Section):
    """
    Bloco com informações das usinas pertencentes a cada conjunto
    de restrições elétricas por conjunto de RE.
    """

    FIM_BLOCO = "999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_conjunto: List[Field] = [IntegerField(3, 0)]
        campos_usinas: List[Field] = [
            IntegerField(3, 6 + i * 4) for i in range(10)
        ]
        self.__linha = Line(campo_conjunto + campos_usinas)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUsinasConjuntoRE):
            return False
        bloco: BlocoUsinasConjuntoRE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = ["Conjunto"] + [f"Usina {i}" for i in range(1, 11)]
            df = pd.DataFrame(tabela, columns=cols)
            df = df.astype({"Conjunto": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_RES, 11))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoUsinasConjuntoRE.FIM_BLOCO in linha:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                tabela[i, :] = dados
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do re.dat não foram lidos com sucesso")

        for _, linha in self.data.iterrows():
            linha_escrita = []
            for v in linha:
                linha_escrita.append(None if np.isnan(v) else int(v))
            file.write(self.__linha.write(linha_escrita))
        file.write(BlocoUsinasConjuntoRE.FIM_BLOCO + "\n")


class BlocoConfiguracaoRestricoesRE(Section):
    """
    Bloco com informações de configuração das restrições elétricas
    para cada conjunto de usinas.
    """

    FIM_BLOCO = "999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 0),
                IntegerField(2, 4),
                IntegerField(4, 7),
                IntegerField(2, 12),
                IntegerField(4, 15),
                IntegerField(1, 20),
                FloatField(15, 22, 2),
                LiteralField(23, 38),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoConfiguracaoRestricoesRE):
            return False
        bloco: BlocoConfiguracaoRestricoesRE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = [
                "Conjunto",
                "Mês Início",
                "Ano Início",
                "Mês Fim",
                "Ano Fim",
                "Flag P",
                "Restrição",
            ]
            df = pd.DataFrame(tabela, columns=cols)
            df["Motivo"] = motivos
            df = df.astype(
                {
                    "Conjunto": "int64",
                    "Mês Início": "int64",
                    "Ano Início": "int64",
                    "Mês Fim": "int64",
                    "Ano Fim": "int64",
                    "Flag P": "int64",
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        motivos: List[str] = []
        tabela = np.zeros((MAX_RES, 7))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoUsinasConjuntoRE.FIM_BLOCO in linha:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                tabela[i, :] = dados[:-1]
                motivos.append(dados[-1])
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do re.dat não foram lidos com sucesso")

        for _, dados_linhas in self.data.iterrows():
            file.write(self.__linha.write(dados_linhas))
        file.write(BlocoUsinasConjuntoRE.FIM_BLOCO + "\n")



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.datetimefield import DatetimeField
from datetime import datetime


class RegistroEolicaConfiguracao(Register):
    """
    Registro que contém um estado de operação de uma usina por
    período.
    """

    IDENTIFIER = "EOLICA-CONFIGURACAO-PERIODO"
    IDENTIFIER_DIGITS = 27
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            LiteralField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def data_inicial_estado_operacao(self) -> Optional[datetime]:
        return self.data[1]

    @data_inicial_estado_operacao.setter
    def data_inicial_estado_operacao(self, n: datetime):
        self.data[1] = n

    @property
    def data_final_estado_operacao(self) -> Optional[datetime]:
        return self.data[2]

    @data_final_estado_operacao.setter
    def data_final_estado_operacao(self, n: datetime):
        self.data[2] = n

    @property
    def estado_operacao(self) -> Optional[str]:
        return self.data[3]

    @estado_operacao.setter
    def estado_operacao(self, n: str):
        self.data[3] = n



from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from typing import List


class RegistroVazoesPostos(Register):
    """
    Registro com os dados associados às vazões dos postos.
    """

    POSTOS = 320

    LINE = Line(
        [IntegerField(size=4, starting_position=4 * i) for i in range(POSTOS)],
        storage="BINARY",
    )

    @classmethod
    def set_postos(cls, postos: int):
        cls.POSTOS = postos
        cls.LINE = Line(
            [
                IntegerField(size=4, starting_position=4 * i)
                for i in range(postos)
            ],
            storage="BINARY",
        )

    @property
    def vazoes(self) -> List[int]:
        return self.data

    @vazoes.setter
    def vazoes(self, v: List[int]):
        self.data = v



from inewave.nwlistop.dlpptbmaxs import DLPPtbmaxs

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.dlpptbmaxs import MockDLPPtbmaxs


def test_atributos_encontrados_dlpptbmaxs():
    m: MagicMock = mock_open(read_data="".join(MockDLPPtbmaxs))
    with patch("builtins.open", m):
        n = DLPPtbmaxs.le_arquivo("")
        assert n.valores is not None
        assert n.valores.iloc[0, 0] == 2020
        assert n.valores.iloc[-1, -1] == 0.0


def test_atributos_nao_encontrados_dlpptbmaxs():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = DLPPtbmaxs.le_arquivo("")
        assert n.valores is None


def test_eq_dlpptbmaxs():
    m: MagicMock = mock_open(read_data="".join(MockDLPPtbmaxs))
    with patch("builtins.open", m):
        n1 = DLPPtbmaxs.le_arquivo("")
        n2 = DLPPtbmaxs.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from inewave.nwlistop.modelos.earmfsin import EarmAnos

from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class EarmfSIN(ArquivoSIN):
    """
    Armazena os dados das saídas referentes à energia armazenada
    final em MWmes para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `eafmfsin.out`.
    """

    BLOCKS = [
        EarmAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="earmfsin.out"
    ) -> "EarmfSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="earmfsin.out"):
        self.write(diretorio, nome_arquivo)



import pandas as pd  # type: ignore

from inewave.nwlistop.modelos.mediasmerc import LeituraMediasMerc


class MediasMerc:
    """
    Armazena os dados das saídas referentes às médias de diversas variáveis
    agrupadas por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `MEDIAS-MERC.CSV`.

    """

    def __init__(self, dados: pd.DataFrame):
        self.__dados = dados

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre MediasMerc avalia todos os valores da tabela.
        """
        if not isinstance(o, MediasMerc):
            return False
        m: MediasMerc = o
        return self.medias.equals(m.medias)

    @property
    def medias(self) -> pd.DataFrame:
        return self.__dados

    @medias.setter
    def medias(self, d: pd.DataFrame) -> pd.DataFrame:
        self.__dados = d

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="MEDIAS-MERC.CSV"):
        return cls(LeituraMediasMerc(diretorio).le_arquivo(nome_arquivo))
